# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import platform
import shutil
import tempfile

from abc import ABC, abstractmethod
from typing import Final, List, Mapping, NamedTuple, Sequence

from pydub import AudioSegment
from pydub.effects import speedup

_DEFAULT_CHUNK_SIZE: Final[int] = 150


class Voice(NamedTuple):
    name: str
    gender: str
    region: str = ""


class TextToSpeech(ABC):

    def __init__(self):
        self._SSML_MALE: Final[str] = "Male"
        self._SSML_FEMALE: Final[str] = "Female"
        self._DEFAULT_SPEED: Final[float] = 1.0
        self._DEFAULT_VOLUME_GAIN_DB: Final[float] = 16.0

    @abstractmethod
    def get_available_voices(self, language_code: str) -> List[Voice]:
        pass

    def get_voices_with_region_preference(
        self, *, voices: List[Voice], target_language_region: str
    ) -> List[Voice]:
        if len(target_language_region) == 0:
            return voices

        voices_copy = voices[:]

        for voice in voices:
            if voice.region.endswith(target_language_region):
                voices_copy.remove(voice)
                voices_copy.insert(0, voice)

        return voices_copy

    def assign_voices(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        target_language: str,
        target_language_region: str,
    ) -> Mapping[str, str | None]:

        voices = self.get_available_voices(target_language)
        voices = self.get_voices_with_region_preference(
            voices=voices, target_language_region=target_language_region
        )

        voice_assignment = {}
        if len(voices) == 0:
            voice_assignment["speaker_01"] = "ona"
        else:
            for chunk in utterance_metadata:
                speaker_id = chunk["speaker_id"]
                if speaker_id in voice_assignment:
                    continue

                gender = chunk["ssml_gender"]
                for voice in voices:
                    if voice.gender.lower() == gender.lower():
                        voice_assignment[speaker_id] = voice.name
                        break

        logging.debug(f"text_to_speech.assign_voices. Returns: {voice_assignment}")
        return voice_assignment

    def _convert_to_mp3(self, input_file, output_mp3):
        null_device = "NUL" if platform.system().lower() == "windows" else "/dev/null"
        cmd = f"ffmpeg -y -i {input_file} {output_mp3} > {null_device} 2>&1"
        logging.debug(cmd)
        os.system(cmd)
        os.remove(input_file)

    def _add_text_to_speech_properties(
        self,
        *,
        utterance_metadata: Mapping[str, str | float],
    ) -> Mapping[str, str | float]:
        """Updates utterance metadata with Text-To-Speech properties."""
        utterance_metadata_copy = utterance_metadata.copy()
        voice_properties = dict(
            pitch=0,
            speed=self._DEFAULT_SPEED,
            volume_gain_db=self._DEFAULT_VOLUME_GAIN_DB,
        )
        utterance_metadata_copy.update(voice_properties)
        return utterance_metadata_copy

    def update_utterance_metadata(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        assigned_voices: Mapping[str, str] | None,
    ) -> Sequence[Mapping[str, str | float]]:
        """Updates utterance metadata with assigned voices."""
        updated_utterance_metadata = []
        for metadata_item in utterance_metadata:
            new_utterance = metadata_item.copy()
            speaker_id = new_utterance.get("speaker_id")
            new_utterance["assigned_voice"] = assigned_voices.get(speaker_id)
            new_utterance = self._add_text_to_speech_properties(
                utterance_metadata=new_utterance
            )
            updated_utterance_metadata.append(new_utterance)
        return updated_utterance_metadata

    @abstractmethod
    def get_languages(self):
        pass

    """ TTS add silence at the end that we want to remove to prevent increasing the speech of next
        segments if is not necessary."""

    def _convert_text_to_speech_without_end_silence(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        pitch: float,
        speed: float,
        volume_gain_db: float,
    ) -> str:

        dubbed_file = self._convert_text_to_speech(
            assigned_voice=assigned_voice,
            target_language=target_language,
            output_filename=output_filename,
            text=text,
            pitch=pitch,
            speed=speed,
            volume_gain_db=volume_gain_db,
        )

        dubbed_audio = AudioSegment.from_file(dubbed_file)
        pre_duration = len(dubbed_audio)

        filename = ""
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            shutil.copyfile(dubbed_file, temp_file.name)
            null_device = (
                "NUL" if platform.system().lower() == "windows" else "/dev/null"
            )
            cmd = f"ffmpeg -y -i {temp_file.name} -af silenceremove=stop_periods=-1:stop_duration=0.1:stop_threshold=-50dB {dubbed_file} > {null_device} 2>&1"
            os.system(cmd)
            filename = temp_file.name

        if os.path.exists(filename):
            os.remove(filename)

        dubbed_audio = AudioSegment.from_file(dubbed_file)
        post_duration = len(dubbed_audio)
        if pre_duration != post_duration:
            logging.debug(
                f"text_to_speech._convert_text_to_speech_without_end_silence. File {dubbed_file} shorten from {pre_duration} to {post_duration}"
            )

        return dubbed_file

    @abstractmethod
    def _convert_text_to_speech(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        pitch: float,
        speed: float,
        volume_gain_db: float,
    ) -> str:
        pass

    def _calculate_target_utterance_speed(
        self,
        *,
        reference_length: float,
        dubbed_file: str,
    ) -> float:
        """Returns the ratio between the reference and target duration."""

        dubbed_audio = AudioSegment.from_file(dubbed_file)
        dubbed_duration = dubbed_audio.duration_seconds
        r = dubbed_duration / reference_length
        logging.debug(f"text_to_speech._calculate_target_utterance_speed: {r}")
        return r

    def create_speaker_to_paths_mapping(
        self,
        utterance_metadata: Sequence[Mapping[str, float | str]],
    ) -> Mapping[str, Sequence[str]]:
        """Organizes a list of utterance metadata dictionaries into a speaker-to-paths mapping.

        Returns:
            A mapping between speaker IDs to lists of file paths.
        """

        speaker_to_paths_mapping = {}
        for utterance in utterance_metadata:
            speaker_id = utterance["speaker_id"]
            if speaker_id not in speaker_to_paths_mapping:
                speaker_to_paths_mapping[speaker_id] = []
            speaker_to_paths_mapping[speaker_id].append(utterance["vocals_path"])
        return speaker_to_paths_mapping

    def _adjust_audio_speed(
        self,
        *,
        reference_length: float,
        dubbed_file: str,
        speed: float,
        chunk_size: int = _DEFAULT_CHUNK_SIZE,
    ) -> None:
        """Adjusts the speed of an MP3 file to match the reference file duration.

        The speed will not be adjusted if the dubbed file has a duration that
        is the same or shorter than the duration of the reference file.
        """

        dubbed_audio = AudioSegment.from_file(dubbed_file)
        logging.info(
            "Adjusting audio speed will prevent overlaps of utterances. However,"
            " it might change the voice sligthly."
        )
        crossfade = max(1, chunk_size // 2)

        logging.debug(
            f"text_to_speech.adjust_audio_speed: dubbed_audio: {dubbed_file}, speed: {speed}, chunk_size: {chunk_size}, crossfade: {crossfade}"
        )
        output_audio = speedup(
            dubbed_audio, speed, chunk_size=chunk_size, crossfade=crossfade
        )
        output_audio.export(dubbed_file, format="mp3")

    def _does_voice_supports_speeds(self):
        return False

    def get_start_time_of_next_speech_utterance(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        from_time: float,
    ) -> int:
        result = None
        for utterance in utterance_metadata:
            start = utterance["start"]
            if start <= from_time:
                continue

            for_dubbing = utterance["for_dubbing"]
            if not for_dubbing:
                continue

            result = start
            break

        logging.debug(
            f"get_start_time_of_next_speech_utterance from_time:{from_time}, result: {result}"
        )
        return result

    def _do_need_to_increase_speed(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        dubbed_path: str,
        start: float,
    ) -> bool:

        next_start = self.get_start_time_of_next_speech_utterance(
            utterance_metadata=utterance_metadata, from_time=start
        )
        dubbed_audio = AudioSegment.from_file(dubbed_path)
        dubbed_duration = dubbed_audio.duration_seconds
        end = dubbed_duration + start
        logging.debug(
            f"_do_need_to_increase_speed. start: {start}, next_start: {next_start} < end: {end}, duration: {dubbed_duration}"
        )
        if next_start and end < next_start:
            return False

        return True

    def dub_utterances(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        output_directory: str,
        target_language: str,
        adjust_speed: bool = True,
    ) -> Sequence[Mapping[str, str | float]]:
        """Processes a list of utterance metadata, generating dubbed audio files."""

        logging.debug(f"TextToSpeech.dub_utterances: adjust_speed: {adjust_speed}")
        updated_utterance_metadata = []
        for utterance in utterance_metadata:
            utterance_copy = utterance.copy()
            if not utterance_copy["for_dubbing"]:
                try:
                    dubbed_path = utterance_copy["path"]
                except KeyError:
                    dubbed_path = f"chunk_{utterance['start']}_{utterance['end']}.mp3"
            else:
                assigned_voice = utterance_copy["assigned_voice"]
                reference_length = utterance_copy["end"] - utterance_copy["start"]
                text = utterance_copy["translated_text"]
                try:
                    path = utterance_copy["path"]
                    base_filename = os.path.splitext(os.path.basename(path))[0]
                    output_filename = os.path.join(
                        output_directory, f"dubbed_{base_filename}.mp3"
                    )
                except KeyError:
                    output_filename = os.path.join(
                        output_directory,
                        f"dubbed_chunk_{utterance['start']}_{utterance['end']}.mp3",
                    )

                speed = utterance_copy["speed"]
                dubbed_path = self._convert_text_to_speech_without_end_silence(
                    assigned_voice=assigned_voice,
                    target_language=target_language,
                    output_filename=output_filename,
                    text=text,
                    pitch=utterance_copy["pitch"],
                    speed=speed,
                    volume_gain_db=utterance_copy["volume_gain_db"],
                )
                assigned_voice = utterance_copy.get("assigned_voice", None)
                assigned_voice = assigned_voice if assigned_voice else ""
                support_speeds = self._does_voice_supports_speeds()
                speed = self._calculate_target_utterance_speed(
                    reference_length=reference_length, dubbed_file=dubbed_path
                )
                logging.debug(f"support_speeds: {support_speeds}, speed: {speed}")

                if speed > 1.0:
                    increase_speed = self._do_need_to_increase_speed(
                        utterance_metadata=utterance_metadata,
                        dubbed_path=dubbed_path,
                        start=utterance_copy["start"],
                    )
                    translated_text = utterance_copy["translated_text"]
                    if not increase_speed:
                        logging.debug(
                            f"text_to_speech.dub_utterances. No need to increase speed for '{translated_text}'"
                        )
                    else:
                        logging.debug(
                            f"text_to_speech.dub_utterances. Need to increase speed for '{translated_text}'"
                        )

                else:
                    increase_speed = False

                if increase_speed and speed > 1.0:
                    MAX_SPEED = 1.3
                    if speed > MAX_SPEED:
                        logging.debug(
                            f"text_to_speech.dub_utterances: Reduced speed from {speed} to {MAX_SPEED}"
                        )
                        speed = MAX_SPEED

                    translated_text = utterance_copy["translated_text"]
                    logging.debug(
                        f"text_to_speech.dub_utterances: Adjusting speed to {speed} for '{translated_text}'"
                    )

                    utterance_copy["speed"] = speed
                    if support_speeds:
                        dubbed_path = self._convert_text_to_speech_without_end_silence(
                            assigned_voice=assigned_voice,
                            target_language=target_language,
                            output_filename=output_filename,
                            text=text,
                            pitch=utterance_copy["pitch"],
                            speed=speed,
                            volume_gain_db=utterance_copy["volume_gain_db"],
                        )
                    else:
                        chunk_size = utterance_copy.get(
                            "chunk_size", _DEFAULT_CHUNK_SIZE
                        )
                        self._adjust_audio_speed(
                            reference_length=reference_length,
                            dubbed_file=dubbed_path,
                            speed=speed,
                            chunk_size=chunk_size,
                        )
                        utterance_copy["chunk_size"] = chunk_size

            utterance_copy["dubbed_path"] = dubbed_path
            updated_utterance_metadata.append(utterance_copy)
        return updated_utterance_metadata



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from transformers import AutoModelForSeq2SeqLM, AutoTokenizer, pipeline

from open_dubbing.translation import Translation


class TranslationNLLB(Translation):

    def __init__(self, device="cpu"):
        super().__init__(device)
        self.translator = None
        self.translator_languages = ""

    def load_model(self, name="nllb-200-1.3B"):
        self.model_name = f"facebook/{name}"
        self.tokenizer = self._get_tokenizer_nllb()

    def _translate_text(
        self, source_language: str, target_language: str, text: str
    ) -> str:
        languages = f"{source_language}{target_language}"
        if not self.translator or self.translator_languages != languages:
            model = self._get_model_nllb()
            self.translator = pipeline(
                "translation",
                model=model,
                tokenizer=self.tokenizer,
                src_lang=self._get_nllb_language(source_language),
                tgt_lang=self._get_nllb_language(target_language),
                max_length=1024,
            )
            self.translator_languages = languages

        translated = self.translator(text)
        return translated[0]["translation_text"]

    def _get_tokenizer_nllb(self):
        return AutoTokenizer.from_pretrained(self.model_name)

    def _get_model_nllb(self):
        try:
            return AutoModelForSeq2SeqLM.from_pretrained(self.model_name).to(
                self.device
            )
        except RuntimeError as e:
            if self.device == "cuda":
                return AutoModelForSeq2SeqLM.from_pretrained(self.model_name).to("cpu")
                logging.warning(
                    f"Loading translation model {self.model_name} in CPU since cannot be load in GPU"
                )
            else:
                raise e

    def get_language_pairs(self):
        tokenizer = self._get_tokenizer_nllb()
        # Returns 'cat_Latn'
        original_list = tokenizer.additional_special_tokens
        # Get only the language codes
        supported_languages = [s[:3] for s in original_list]
        pairs = set()
        for source in supported_languages:
            for target in supported_languages:
                if source == target:
                    continue

                pair = (source, target)
                pairs.add(pair)

        return pairs

    def _get_nllb_language(self, source_language_iso_639_3: str) -> str:
        tokenizer = self._get_tokenizer_nllb()
        nllb_languages = tokenizer.additional_special_tokens
        for nllb_language in nllb_languages:
            if nllb_language[:3] == source_language_iso_639_3:
                return nllb_language

        raise ValueError(
            f"Language {source_language_iso_639_3} not supported by Meta NLLB translation model"
        )



# Copyright 2024 Jordi Mas i Herǹadez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os

from typing import List

from open_dubbing.text_to_speech import TextToSpeech, Voice


class TextToSpeechCLI(TextToSpeech):

    def __init__(self, device="cpu", configuration_file=None):
        super().__init__()
        self.device = device
        self.configuration = self.load_json(configuration_file)
        self.output_dir = os.path.abspath("tts-ouput/")

    def load_json(self, configuration_file):
        with open(configuration_file, "r") as file:
            data = json.load(file)

        return data

    def get_available_voices(self, language_code: str) -> List[Voice]:
        voices = []

        for voice_cfg in self.configuration["voices"]:
            if voice_cfg["language"] != language_code:
                continue

            voice = Voice(
                name=voice_cfg["id"],
                gender=voice_cfg["gender"],
                region=voice_cfg["region"],
            )
            voices.append(voice)

        logging.debug(
            f"text_to_speech_cli.get_available_voices: {voices} for language {language_code}"
        )
        return voices

    def _get_command(self, *, assigned_voice: str, directory: str, text: str):
        command = self.configuration["command"]
        cmd = command.format(
            assigned_voice=assigned_voice,
            text=text,
            directory=directory,
            device=self.device,
        )
        return cmd

    def _get_output_pattern(self, *, assigned_voice: str, directory: str, text: str):

        output_pattern = self.configuration["output_pattern"]
        pattern = output_pattern.format(
            assigned_voice=assigned_voice, text=text, directory=directory
        )
        return pattern

    def _convert_text_to_speech(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        pitch: float,
        speed: float,
        volume_gain_db: float,
    ) -> str:

        cmd = self._get_command(
            assigned_voice=assigned_voice, text=text, directory=self.output_dir
        )
        return_code = os.system(cmd)
        if return_code != 0:
            raise RuntimeError(
                f"Command '{cmd}' failed with return code: {return_code}"
            )

        wav_file = self._get_output_pattern(
            assigned_voice=assigned_voice,
            text=text,
            directory=self.output_dir,
        )
        self._convert_to_mp3(wav_file, output_filename)

        logging.debug(f"text_to_speech_cli._convert_text_to_speech: {text}")
        return output_filename

    def get_languages(self):
        languages = set()
        for voice_cfg in self.configuration["voices"]:
            language = voice_cfg["language"]
            languages.add(language)

        languages_list = list(languages)

        logging.debug(f"text_to_speech_cli.get_languages: {languages_list}")
        return languages_list



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys

from iso639 import Lang

from open_dubbing.command_line import CommandLine
from open_dubbing.dubbing import Dubber
from open_dubbing.speech_to_text_faster_whisper import SpeechToTextFasterWhisper
from open_dubbing.speech_to_text_whisper_transformers import (
    SpeechToTextWhisperTransfomers,
)
from open_dubbing.text_to_speech_cli import TextToSpeechCLI
from open_dubbing.text_to_speech_edge import TextToSpeechEdge
from open_dubbing.text_to_speech_mms import TextToSpeechMMS
from open_dubbing.translation_apertium import TranslationApertium
from open_dubbing.translation_nllb import TranslationNLLB
from open_dubbing.video_processing import VideoProcessing


def _init_logging(log_level):
    # Create a logger
    logger = logging.getLogger()
    logger.setLevel(log_level)  # Set the global log level

    # File handler for logging to a file
    file_handler = logging.FileHandler("open_dubbing.log")
    console_handler = logging.StreamHandler()

    # Formatter for log messages
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

    # Set formatter for both handlers
    file_handler.setFormatter(formatter)
    console_handler.setFormatter(formatter)

    # Add handlers to the logger
    logger.addHandler(file_handler)
    logger.addHandler(console_handler)

    logging.getLogger("pydub.converter").setLevel(logging.ERROR)


def check_languages(source_language, target_language, _tts, translation, _sst):
    spt = _sst.get_languages()
    translation_languages = translation.get_language_pairs()
    logging.debug(f"check_languages. Pairs {len(translation_languages)}")

    tts = _tts.get_languages()

    if source_language not in spt:
        raise ValueError(
            f"source language '{source_language}' is not supported by the speech recognition system. Supported languages: '{spt}"
        )

    pair = (source_language, target_language)
    if pair not in translation_languages:
        raise ValueError(
            f"language pair '{pair}' is not supported by the translation system."
        )

    if target_language not in tts:
        raise ValueError(
            f"target language '{target_language}' is not supported by the text to speech system. Supported languages: '{tts}"
        )


_ACCEPTED_VIDEO_FORMATS = ["mp4"]


def check_is_a_video(input_file: str):
    _, file_extension = os.path.splitext(input_file)
    file_extension = file_extension.lower().lstrip(".")

    if file_extension in _ACCEPTED_VIDEO_FORMATS:
        return
    raise ValueError(f"Unsupported file format: {file_extension}")


HUGGING_FACE_VARNAME = "HF_TOKEN"


def get_token(provided_token: str) -> str:
    token = provided_token or os.getenv(HUGGING_FACE_VARNAME)
    if not token:
        raise ValueError(
            f"You must either provide the '--hugging_face_token' argument or"
            f" set the '{HUGGING_FACE_VARNAME.upper()}' environment variable."
        )
    return token


def _get_language_names(languages_iso_639_3):
    names = []
    for language in languages_iso_639_3:
        o = Lang(language)
        names.append(o.name)
    return sorted(names)


def list_supported_languages(_tts, translation, device):  # TODO: Not used
    s = SpeechToTextFasterWhisper(device=device)
    s.load_model()
    spt = s.get_languages()
    trans = translation.get_languages()
    tts = _tts.get_languages()

    source = _get_language_names(set(spt).intersection(set(trans)))
    print(f"Supported source languages: {source}")

    target = _get_language_names(set(tts).intersection(set(trans)))
    print(f"Supported target languages: {target}")


def main():

    args = CommandLine.read_parameters()
    _init_logging(args.log_level)

    check_is_a_video(args.input_file)

    hugging_face_token = get_token(args.hugging_face_token)

    if not VideoProcessing.is_ffmpeg_installed():
        raise ValueError("You need to have ffmpeg (which includes ffprobe) installed.")

    if args.tts == "mms":
        tts = TextToSpeechMMS(args.device)
    elif args.tts == "edge":
        tts = TextToSpeechEdge(args.device)
    elif args.tts == "coqui":
        try:
            from open_dubbing.coqui import Coqui
            from open_dubbing.text_to_speech_coqui import TextToSpeechCoqui
        except Exception:
            raise ValueError(
                "Make sure that Coqui-tts is installed by running 'pip install open-dubbing[coqui]'"
            )

        tts = TextToSpeechCoqui(args.device)
        if not Coqui.is_espeak_ng_installed():
            raise ValueError(
                "To use Coqui-tts you have to have espeak or espeak-ng installed"
            )
    elif args.tts == "cli":
        if len(args.tts_cli_cfg_file) == 0:
            raise ValueError(
                "When using the tts CLI you need to provide a configuration file which describes the commands and voices to use."
            )

        tts = TextToSpeechCLI(args.device, args.tts_cli_cfg_file)
    else:
        raise ValueError(f"Invalid tts value {args.tts}")

    if sys.platform == "darwin":
        os.environ["TOKENIZERS_PARALLELISM"] = "false"

    if args.stt == "auto":
        if sys.platform == "darwin":
            stt = SpeechToTextWhisperTransfomers(
                model_name=args.whisper_model,
                device=args.device,
                cpu_threads=args.cpu_threads,
            )
        else:
            stt = SpeechToTextFasterWhisper(
                model_name=args.whisper_model,
                device=args.device,
                cpu_threads=args.cpu_threads,
            )
    elif args.stt == "faster-whisper":
        stt = SpeechToTextFasterWhisper(
            model_name=args.whisper_model,
            device=args.device,
            cpu_threads=args.cpu_threads,
        )
    else:
        stt = SpeechToTextWhisperTransfomers(
            model_name=args.whisper_model,
            device=args.device,
            cpu_threads=args.cpu_threads,
        )

    stt.load_model()
    source_language = args.source_language
    if not source_language:
        source_language = stt.detect_language(args.input_file)
        logging.info(f"Detected language '{source_language}'")

    if args.translator == "nllb":
        translation = TranslationNLLB(args.device)
        translation.load_model(args.nllb_model)
    elif args.translator == "apertium":
        server = args.apertium_server
        if len(server) == 0:
            raise ValueError(
                "When using Apertium's API, you need to specify with --apertium-server the URL of the server"
            )

        translation = TranslationApertium(args.device)
        translation.set_server(server)
    else:
        raise ValueError(f"Invalid translator value {args.translator}")

    check_languages(source_language, args.target_language, tts, translation, stt)

    if not os.path.exists(args.output_directory):
        os.makedirs(args.output_directory)

    dubber = Dubber(
        input_file=args.input_file,
        output_directory=args.output_directory,
        source_language=source_language,
        target_language=args.target_language,
        target_language_region=args.target_language_region,
        hugging_face_token=hugging_face_token,
        tts=tts,
        translation=translation,
        stt=stt,
        device=args.device,
        cpu_threads=args.cpu_threads,
        debug=args.debug,
    )
    logging.info(
        f"Processing '{args.input_file}' file with tts '{args.tts}', sst {args.stt} and device '{args.device}'"
    )
    dubber.dub()


if __name__ == "__main__":
    main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from typing import Final, Mapping, Sequence

import torch

from pyannote.audio import Pipeline
from pydub import AudioSegment

_DEFAULT_DUBBED_VOCALS_AUDIO_FILE: Final[str] = "dubbed_vocals.mp3"
_DEFAULT_DUBBED_AUDIO_FILE: Final[str] = "dubbed_audio"
_DEFAULT_OUTPUT_FORMAT: Final[str] = ".mp3"
_SUPPORTED_DEVICES: Final[tuple[str, str]] = ("cpu", "cuda")
_TIMESTAMP_THRESHOLD: Final[float] = 0.001


def create_pyannote_timestamps(
    *,
    audio_file: str,
    pipeline: Pipeline,
    device: str = "cpu",
) -> Sequence[Mapping[str, float]]:
    """Creates timestamps from a vocals file using Pyannote speaker diarization.

    Returns:
        A list of dictionaries containing start and end timestamps for each
        speaker segment.
    """
    if device not in _SUPPORTED_DEVICES:
        raise ValueError(
            "The device must be either (' or ').join(_SUPPORTED_DEVICES). Got:"
            f" {device}"
        )
    if device == "cuda":
        pipeline.to(torch.device("cuda"))
    diarization = pipeline(audio_file)
    utterance_metadata = [
        {"start": segment.start, "end": segment.end, "speaker_id": speaker}
        for segment, _, speaker in diarization.itertracks(yield_label=True)
    ]
    return utterance_metadata


def _cut_and_save_audio(
    *,
    audio: AudioSegment,
    utterance: Mapping[str, str | float],
    prefix: str,
    output_directory: str,
) -> str:
    """Cuts a specified segment from an audio file, saves it as an MP3, and returns the path of the saved file.

    Args:
        audio: The audio file from which to extract the segment.
        utterance: A dictionary containing the start and end times of the segment
          to be cut. - 'start': The start time of the segment in seconds. - 'end':
          The end time of the segment in seconds.
        prefix: A string to be used as a prefix in the filename of the saved audio
          segment.
        output_directory: The directory path where the cut audio segment will be
          saved.

    Returns:
        The path of the saved MP3 file.
    """
    start_time_ms = int(utterance["start"] * 1000)
    end_time_ms = int(utterance["end"] * 1000)
    chunk = audio[start_time_ms:end_time_ms]
    chunk_filename = f"{prefix}_{utterance['start']}_{utterance['end']}.mp3"
    chunk_path = f"{output_directory}/{chunk_filename}"
    chunk.export(chunk_path, format="mp3")
    return chunk_path


def run_cut_and_save_audio(
    *,
    utterance_metadata: Sequence[Mapping[str, float]],
    audio_file: str,
    output_directory: str,
) -> Sequence[Mapping[str, float]]:
    """Cuts an audio file into chunks based on provided time ranges and saves each chunk to a file.

    Returns:
        A list of dictionaries, each containing the path to the saved chunk, and
        the original start and end times.
    """

    audio = AudioSegment.from_file(audio_file)
    key = "path"
    prefix = "chunk"
    updated_utterance_metadata = []
    for utterance in utterance_metadata:
        chunk_path = _cut_and_save_audio(
            audio=audio,
            utterance=utterance,
            prefix=prefix,
            output_directory=output_directory,
        )
        utterance_copy = utterance.copy()
        utterance_copy[key] = chunk_path
        updated_utterance_metadata.append(utterance_copy)
    return updated_utterance_metadata


def insert_audio_at_timestamps(
    *,
    utterance_metadata: Sequence[Mapping[str, str | float]],
    background_audio_file: str,
    output_directory: str,
) -> str:
    """Inserts audio chunks into a background audio track at specified timestamps."""

    background_audio = AudioSegment.from_mp3(background_audio_file)
    total_duration = background_audio.duration_seconds
    output_audio = AudioSegment.silent(duration=total_duration * 1000)
    for item in utterance_metadata:
        audio_chunk = AudioSegment.from_mp3(item["dubbed_path"])
        start_time = int(item["start"] * 1000)
        output_audio = output_audio.overlay(
            audio_chunk, position=start_time, loop=False
        )
    dubbed_vocals_audio_file = os.path.join(
        output_directory, _DEFAULT_DUBBED_VOCALS_AUDIO_FILE
    )
    output_audio.export(dubbed_vocals_audio_file, format="mp3")
    return dubbed_vocals_audio_file


def merge_background_and_vocals(
    *,
    background_audio_file: str,
    dubbed_vocals_audio_file: str,
    output_directory: str,
    target_language: str,
    vocals_volume_adjustment: float = 5.0,
    background_volume_adjustment: float = 0.0,
) -> str:
    """Mixes background music and vocals tracks, normalizes the volume, and exports the result.

    Returns:
      The path to the output audio file with merged dubbed vocals and original
      background audio.
    """

    background = AudioSegment.from_mp3(background_audio_file)
    vocals = AudioSegment.from_mp3(dubbed_vocals_audio_file)
    background = background.normalize()
    vocals = vocals.normalize()
    background = background + background_volume_adjustment
    vocals = vocals + vocals_volume_adjustment
    shortest_length = min(len(background), len(vocals))
    background = background[:shortest_length]
    vocals = vocals[:shortest_length]
    mixed_audio = background.overlay(vocals)
    target_language_suffix = "_" + target_language.replace("-", "_").lower()
    dubbed_audio_file = os.path.join(
        output_directory,
        _DEFAULT_DUBBED_AUDIO_FILE + target_language_suffix + _DEFAULT_OUTPUT_FORMAT,
    )
    mixed_audio.export(dubbed_audio_file, format="mp3")
    return dubbed_audio_file



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import logging

import numpy as np

from faster_whisper import WhisperModel

from open_dubbing.speech_to_text import SpeechToText


class SpeechToTextFasterWhisper(SpeechToText):

    def __init__(self, *, model_name="medium", device="cpu", cpu_threads=0):
        super().__init__(device=device, model_name=model_name, cpu_threads=cpu_threads)

        logging.getLogger("faster_whisper").setLevel(logging.ERROR)

    def load_model(self):
        self._model = WhisperModel(
            model_size_or_path=self.model_name,
            device=self.device,
            cpu_threads=self.cpu_threads,
            compute_type="float16" if self.device == "cuda" else "int8",
        )

    def get_languages(self):
        iso_639_3 = []
        for language in self.model.supported_languages:
            pt3 = self._get_iso_639_3(language)
            iso_639_3.append(pt3)
        return iso_639_3

    def _transcribe(
        self,
        *,
        vocals_filepath: str,
        source_language_iso_639_1: str,
    ) -> str:
        segments, _ = self.model.transcribe(
            vocals_filepath,
            source_language_iso_639_1,
        )
        return " ".join(segment.text for segment in segments)

    def _get_audio_language(self, audio: array.array) -> str:
        audio_input = np.array(audio).astype(np.float32) / 32768.0
        _, info = self.model.transcribe(audio_input)
        detected_language = self._get_iso_639_3(info.language)
        logging.debug(
            f"speech_to_text_faster_whisper._get_audio_language. Detected language: {detected_language}"
        )
        return detected_language



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import logging

from abc import ABC, abstractmethod
from typing import Mapping, Sequence

from iso639 import Lang
from pydub import AudioSegment

from open_dubbing.voice_gender_classifier import VoiceGenderClassifier


class SpeechToText(ABC):

    def __init__(self, *, model_name="medium", device="cpu", cpu_threads=0):
        self.model_name = model_name
        self.model = None
        self.device = device
        self.cpu_threads = cpu_threads
        self.MIN_SECS = 0.5

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @abstractmethod
    def load_model(self):
        pass

    @abstractmethod
    def get_languages(self):
        pass

    def _get_iso_639_1(self, iso_639_3: str):
        o = Lang(iso_639_3)
        iso_639_1 = o.pt1
        return iso_639_1

    def _get_iso_639_3(self, iso_639_1: str):
        if iso_639_1 == "jw":
            iso_639_1 = "jv"

        o = Lang(iso_639_1)
        iso_639_3 = o.pt3
        return iso_639_3

    @abstractmethod
    def _transcribe(
        self,
        *,
        vocals_filepath: str,
        source_language_iso_639_1: str,
    ) -> str:
        pass

    def transcribe_audio_chunks(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, float | str]],
        source_language: str,
        no_dubbing_phrases: Sequence[str],
    ) -> Sequence[Mapping[str, float | str]]:

        logging.debug(f"transcribe_audio_chunks: {source_language}")
        iso_639_1 = self._get_iso_639_1(source_language)

        updated_utterance_metadata = []
        for item in utterance_metadata:
            new_item = item.copy()
            path = ""
            try:
                path = item["path"]
                duration = item["end"] - item["start"]
                if self._is_short_audio(duration=duration):
                    transcribed_text = ""
                    logging.warn(
                        f"speech_to_text._is_short_audio. Audio is less than {self.MIN_SECS} second, skipping transcription of '{path}'."
                    )
                else:
                    transcribed_text = self._transcribe(
                        vocals_filepath=path,
                        source_language_iso_639_1=iso_639_1,
                    )
            except Exception as e:
                logging.error(
                    f"speech_to_text.transcribe_audio_chunks. file '{path}', error: '{e}'"
                )
                transcribed_text = ""

            dubbing = len(transcribed_text) > 0
            logging.debug(
                f"transcribe_audio_chunks. text: '{transcribed_text}' - dubbing: {dubbing}"
            )
            new_item["text"] = transcribed_text
            new_item["for_dubbing"] = dubbing
            updated_utterance_metadata.append(new_item)
        return updated_utterance_metadata

    #  Returns a list of unique speakers with the largest audio sample for the speaker
    def _get_unique_speakers_largest_audio(self, utterance_metadata):
        speakers = {}
        for chunk in utterance_metadata:
            speaker = chunk["speaker_id"]
            length = chunk["end"] - chunk["start"]
            dubbed_path = chunk["path"]

            speaker_data = speakers.get(speaker, {})
            save = False
            if len(speaker_data) == 0:
                save = True
            else:
                if length > speaker_data["length"]:
                    save = True

            if save:
                speaker_data["length"] = length
                speaker_data["path"] = dubbed_path
                speakers[speaker] = speaker_data

        speaker_tuple = [(speaker, data["path"]) for speaker, data in speakers.items()]
        logging.debug(
            f"text_to_speech._get_unique_speakers_largest_audio: {speaker_tuple}"
        )
        return speaker_tuple

    def diarize_speakers(
        self,
        *,
        file: str,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        number_of_speakers: int,
    ) -> Sequence[tuple[str, str]]:

        speaker_gender = {}
        classifier = VoiceGenderClassifier(self.device)
        speakers = self._get_unique_speakers_largest_audio(utterance_metadata)
        for speaker, path in speakers:
            gender = classifier.get_gender_for_file(path)
            speaker_gender[speaker] = gender

        r = []
        for chunk in utterance_metadata:
            speaker = chunk["speaker_id"]
            gender = speaker_gender[speaker]
            _tuple = (speaker, gender)
            r.append(_tuple)

        logging.debug(f"text_to_speech.diarize_speakers. Returns: {r}")
        return r

    def add_speaker_info(
        self,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        speaker_info: Sequence[tuple[str, str]],
    ) -> Sequence[Mapping[str, str | float]]:
        if len(utterance_metadata) != len(speaker_info):
            raise Exception(
                "The length of 'utterance_metadata' and 'speaker_info' must be the"
                " same."
            )
        updated_utterance_metadata = []
        for utterance, (speaker_id, ssml_gender) in zip(
            utterance_metadata, speaker_info
        ):
            new_utterance = utterance.copy()
            new_utterance["speaker_id"] = speaker_id
            new_utterance["ssml_gender"] = ssml_gender
            updated_utterance_metadata.append(new_utterance)
        return updated_utterance_metadata

    @abstractmethod
    def _get_audio_language(self, audio: array.array) -> str:
        pass

    def detect_language(self, filename: str) -> str:
        DURATION_SECS = 30
        audio = AudioSegment.from_file(filename)
        audio = audio.set_channels(1)
        audio = audio.set_frame_rate(16000)

        first_seconds = audio[: DURATION_SECS * 1000].get_array_of_samples()
        return self._get_audio_language(first_seconds)

    # To prevent Whisper hallucinations with very short audios
    def _is_short_audio(self, *, duration: float):
        if duration < self.MIN_SECS:
            return True

        return False



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from typing import List

from iso639 import Lang

from open_dubbing.coqui import Coqui
from open_dubbing.text_to_speech import TextToSpeech, Voice


class TextToSpeechCoqui(TextToSpeech):

    def __init__(self, device="cpu"):
        super().__init__()
        self.coqui = Coqui(device)

    def get_languages(self):
        languages = []
        for iso_639_1 in self.coqui.get_languages():
            o = Lang(iso_639_1)
            iso_639_3 = o.pt3
            languages.append(iso_639_3)

        return languages

    def _get_iso_639_1(self, iso_639_3: str):
        o = Lang(iso_639_3)
        iso_639_1 = o.pt1
        return iso_639_1

    def get_available_voices(self, language_code: str) -> List[Voice]:
        voices = []

        if language_code == "cat":
            voices.append(Voice(name="pau", gender="Male"))
            voices.append(Voice(name="ona", gender="Female"))

        logging.debug(
            f"text_to_speech_coqui.get_available_voices: {voices} for language {language_code}"
        )
        return voices

    def _convert_text_to_speech(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        pitch: float,
        speed: float,
        volume_gain_db: float,
    ) -> str:

        wav_file = output_filename.replace(".mp3", ".wav")
        logging.debug(
            f"text_to_speech.client.synthesize_speech: pre synthesize_speech: '{text}', '{target_language}', file: {wav_file}, speed: {speed}, voice: {assigned_voice}"
        )
        iso_639_1 = self._get_iso_639_1(target_language)
        self.coqui.synthesize_speech(
            text, iso_639_1, file_path=wav_file, voice=assigned_voice
        )

        self._convert_to_mp3(wav_file, output_filename)
        logging.debug(
            f"text_to_speech.client.synthesize_speech: output_filename: '{output_filename}'"
        )
        return output_filename
